// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.notification.provider.azure.messageBus;

import com.microsoft.azure.servicebus.MessageHandlerOptions;
import com.microsoft.azure.servicebus.SubscriptionClient;
import com.microsoft.azure.servicebus.management.ManagementClient;
import com.microsoft.azure.servicebus.primitives.ConnectionStringBuilder;
import com.microsoft.azure.servicebus.primitives.ServiceBusException;
import org.opengroup.osdu.azure.cosmosdb.CosmosStore;
import org.opengroup.osdu.azure.partition.PartitionInfoAzure;
import org.opengroup.osdu.azure.partition.PartitionServiceClient;
import org.opengroup.osdu.azure.serviceBusManager.IManagementClientFactory;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.notification.Subscription;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.core.common.notification.ISubscriptionFactory;
import org.opengroup.osdu.core.common.provider.interfaces.ITenantFactory;
import org.opengroup.osdu.notification.provider.azure.messageBus.interfaces.ISubscriptionManager;
import org.opengroup.osdu.notification.provider.azure.messageBus.models.TopicSubscriptions;
import org.opengroup.osdu.notification.provider.azure.util.AzureServiceBusConfig;
import org.opengroup.osdu.notification.provider.azure.util.AzureCosmosProperties;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnExpression;
import org.springframework.stereotype.Component;

import java.time.Duration;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.stream.Collectors;

@Component
@ConditionalOnExpression("${azure.serviceBus.enabled:true} || ${azure.eventGridToServiceBus.enabled:true}")
public class SubscriptionManagerImpl implements ISubscriptionManager {
    private final static Logger LOGGER = LoggerFactory.getLogger(SubscriptionManagerImpl.class);
    @Autowired
    private TopicSubscriptions topicSubscriptions;
    @Autowired
    private SubscriptionClientFactImpl subscriptionClientFactory;
    @Autowired
    private ProcessNotification processNotification;
    @Autowired
    private ITenantFactory tenantFactory;
    @Autowired
    private ISubscriptionFactory subscriptionFactory;
    @Autowired
    private PartitionServiceClient partitionService;
    @Autowired
    private CosmosStore cosmosStore;
    @Autowired
    private AzureCosmosProperties azureCosmosProperties;
    @Autowired
    private AzureServiceBusConfig azureServiceBusConfig;
    @Autowired
    private IManagementClientFactory factory;

    @Override
    public void subscribeNotificationsEvent() {

        List<String> tenantList = tenantFactory.listTenantInfo().stream().map(TenantInfo::getDataPartitionId)
                .collect(Collectors.toList());
        ExecutorService executorService = Executors
                .newFixedThreadPool(Integer.parseUnsignedInt(azureServiceBusConfig.getNThreads()));
        for (String partition : tenantList) {
            try {
                List<Subscription> subscriptionsList = cosmosStore.findAllItems(partition, azureCosmosProperties.cosmosDBName(),
                        azureCosmosProperties.registerSubscriptionContainerName(), Subscription.class);

                ManagementClient managementClient = factory.getManager(partition);
                for (Subscription subscription : subscriptionsList) {
                    // To check if its a not new subscription.
                    if (!this.topicSubscriptions.checkIfNewTopicSubscription(partition, subscription.getTopic(), subscription.getNotificationId())) {
                        // Update existing subscriptions and skip registration
                        this.topicSubscriptions.updateCurrentTopicSubscriptions(partition, subscription.getTopic(), subscription.getNotificationId());

                    } else {

                        /* This check is added if a Cosmos subscription is created but the corresponding service bus
                       subscription is still not created or creation is in progress.We do not register message handler
                       with the subscription client as it will throw entity not found exception and unregistering is not supported.
                       Check if its a new Subscription Client */
                        if (managementClient.topicExists(subscription.getTopic()) && managementClient.subscriptionExists(subscription.getTopic(), subscription.getNotificationId())) {
                            try {
                                SubscriptionClient subscriptionClient = this.subscriptionClientFactory
                                        .getSubscriptionClient(partition, subscription.getTopic(), subscription.getNotificationId());
                                registerMessageHandler(subscriptionClient, executorService);
                                this.topicSubscriptions.updateCurrentTopicSubscriptions(partition, subscription.getTopic(), subscription.getNotificationId());
                            } catch (InterruptedException | ServiceBusException e) {
                                LOGGER.error("Error while creating or registering subscription client {}", e.getMessage(), e);
                            } catch (Exception e) {
                                LOGGER.error("Unknown exception occurred while creating or registering subscription client: ", e);
                            }
                        }
                    }
                }

            } catch (AppException e) {
                LOGGER.error("Error creating Cosmos Client {}", e.getMessage(), e);
            } catch (Exception e) {
                LOGGER.error("An exception occurred while subscribing to Notification Event : ", e);
            }
        }
        this.topicSubscriptions.clearTopicSubscriptions();

    }

    private void registerMessageHandler(SubscriptionClient subscriptionClient, ExecutorService executorService) throws ServiceBusException, InterruptedException {

        MessageHandler messageHandler = new MessageHandler(subscriptionClient, processNotification, azureServiceBusConfig.getSpringApplicationName());
        subscriptionClient.registerMessageHandler(
                messageHandler,
                new MessageHandlerOptions(Integer.parseUnsignedInt(azureServiceBusConfig.getMaxConcurrentCalls()),
                        false,
                        Duration.ofSeconds(Integer.parseUnsignedInt(azureServiceBusConfig.getMaxLockRenewDurationInSeconds())),
                        Duration.ofSeconds(1)
                ),
                executorService);
    }

    @Override
    public void run() {
        try {
            subscribeNotificationsEvent();
        } catch (Exception e) {
            LOGGER.error("An exception occurred while subscription registration process : ", e);
        }
    }
}
