/*
 * Copyright 2017-2020, Schlumberger
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.notification.provider.gcp.pubsub;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.opengroup.osdu.core.common.http.HttpResponse;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.core.common.provider.interfaces.ITenantFactory;
import org.opengroup.osdu.core.gcp.oqm.driver.OqmDriver;
import org.opengroup.osdu.core.gcp.oqm.driver.OqmDriverRuntimeException;
import org.opengroup.osdu.core.gcp.oqm.model.*;
import org.opengroup.osdu.notification.provider.gcp.config.OqmConfigurationProperties;
import org.opengroup.osdu.notification.provider.gcp.pubsub.di.OqmNotificationHandler;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.springframework.beans.factory.config.BeanDefinition.SCOPE_SINGLETON;

/**
 * Runs once on the service start.
 * 1. Fetches oqm for message broker pull subscriptions in interested topics. Creates the service's subscribers in every found subscription.
 * 2. Checks for the "subscriber control topic" and creates if it is  absent.
 * - This topic is a "control channel" between Register and Notification services.
 * - The former sends events on new pull Subscriptions being created, the latter listens for events and creates corresponding Subscribers.
 */
@Slf4j
@Component
@Scope(SCOPE_SINGLETON)
@ConditionalOnProperty(name = "oqmDriver")
@RequiredArgsConstructor
public class OqmSubscriberManager {

    private final String ACKNOWLEDGE = "message acknowledged by client";
    private final String NOT_ACKNOWLEDGE = "message not acknowledged by client";

    //TODO should be externalized to application.properties
    private static final List<OqmTopic> INTERESTED_TOPICS =
            Stream.of("records-changed", "schema-changed", "status-changed", "legaltags_changed")
                    .map(topicName -> OqmTopic.builder().name(topicName).build()).collect(Collectors.toList());

    private static final String INTERESTED_SUBSCRIPTIONS_PREFIX = "de-";

    private static final OqmSubscriptionQuery INTERESTED_SUBSCRIPTIONS_QUERY = OqmSubscriptionQuery.builder()
            .forAnyOfTopics(INTERESTED_TOPICS).namePrefix(INTERESTED_SUBSCRIPTIONS_PREFIX)
            .subscriberable(true).build();

    private final ITenantFactory tenantInfoFactory;
    private final OqmDriver driver;
    private final OqmNotificationHandler notificationHandler;
    private final OqmConfigurationProperties properties;

    private final Long constructDate = System.currentTimeMillis();
    private final String subscriberControlTopicSubscriptionName = "notification-service-" + constructDate;
    private OqmSubscription subscriberControlTopicSubscription = null;


    @PostConstruct
    void postConstruct() {
        log.info("OqmSubscriberManager bean constructed. Provisioning STARTED");

        //Get all Tenant infos
        for (TenantInfo tenantInfo : tenantInfoFactory.listTenantInfo()) {
            log.info("* OqmSubscriberManager on provisioning tenant {}:", tenantInfo.getDataPartitionId());
            //For every Tenant Destination get "subscriberable" Subscriptions
            for (OqmSubscription subscription : getSubscriberableSubscriptions(tenantInfo)) {
                log.info("* * OqmSubscriberManager on provisioning for tenant {}, subscription {}:", tenantInfo.getDataPartitionId(), subscription.getName());

                //Register a Subscriber on every subscription
                registerSubscriber(tenantInfo, subscription);

                log.info("* * OqmSubscriberManager on provisioning for tenant {}, subscription {}: Subscriber REGISTERED.", tenantInfo.getDataPartitionId(), subscription.getName());
            }
            log.info("* OqmSubscriberManager on provisioning tenant {}: COMPLETED.", tenantInfo.getDataPartitionId());
        }

        TenantInfo tenant = tenantInfoFactory.listTenantInfo().stream().findAny()
                .orElseThrow(() -> new OqmDriverRuntimeException(null, "Unable to get any TenantInfo"));

        log.info("* OqmSubscriberManager on check '{}' subscriber control topic existence:", properties.getRegisterSubscriberControlTopicName());
        OqmTopic controlTopic = driver.getTopic(properties.getRegisterSubscriberControlTopicName(), getDestination(tenant)).orElse(null);
        if (controlTopic != null) {
            log.info("* * OqmSubscriberManager: '{}' subscriber control topic exists.", properties.getRegisterSubscriberControlTopicName());
        } else {
            log.info("* * OqmSubscriberManager: '{}' subscriber control topic doesn't exist. Trying to create it:", properties.getRegisterSubscriberControlTopicName());
            driver.createAndGetTopic(properties.getRegisterSubscriberControlTopicName(), getDestination(tenant));
        }

        log.info("* OqmSubscriberManager on registering '{}' subscriber control topic subscription with name '{}'",
                properties.getRegisterSubscriberControlTopicName(), subscriberControlTopicSubscriptionName);

        OqmSubscription subscriptionRequest = OqmSubscription.builder().name(subscriberControlTopicSubscriptionName)
                .topics(Collections.singletonList(controlTopic)).build();

        subscriberControlTopicSubscription = driver.createAndGetSubscription(subscriptionRequest, getDestination(tenant));
        log.info("* OqmSubscriberManager on registering '{}' subscriber control topic subscription with name '{}': Subscription CREATED.",
                properties.getRegisterSubscriberControlTopicName(), subscriberControlTopicSubscriptionName);

        registerControlTopicSubscriber(tenant, subscriberControlTopicSubscription);
        log.info("* OqmSubscriberManager on registering '{}' subscriber control topic subscription with name '{}': Subscriber REGISTERED.",
                properties.getRegisterSubscriberControlTopicName(), subscriberControlTopicSubscriptionName);

        log.info("OqmSubscriberManager bean constructed. Provisioning COMPLETED");
    }

    @PreDestroy
    void onPreDestroy() {
        log.info("OqmSubscriberManager bean on pre-destroy: STARTED");
        if (subscriberControlTopicSubscription != null) {
            TenantInfo tenant = tenantInfoFactory.listTenantInfo().stream().findAny()
                    .orElseThrow(() -> new OqmDriverRuntimeException(null, "Unable to get any TenantInfo"));
            log.info("* OqmSubscriberManager on delete '{}' subscriber control topic subscription with name '{}': Subscriber REGISTERED.",
                    properties.getRegisterSubscriberControlTopicName(), subscriberControlTopicSubscriptionName);
            driver.deleteSubscription(subscriberControlTopicSubscriptionName, getDestination(tenant));
        }
        log.info("OqmSubscriberManager bean on pre-destroy: COMPLETED");
    }

    private void registerSubscriber(TenantInfo tenantInfo, OqmSubscription subscription) {
        OqmDestination destination = getDestination(tenantInfo);

        OqmMessageReceiver receiver = (oqmMessage, oqmAckReplier) -> {

            String pubsubMessage = oqmMessage.getData();
            String notificationId = subscription.getName();
            Map<String, String> headerAttributes = oqmMessage.getAttributes();


            HttpResponse response;
            boolean ackedNacked = false;
            try {
                response = notificationHandler.notifySubscriber(notificationId, pubsubMessage, headerAttributes);

                if (!response.isSuccessCode()) {
                    log.error(NOT_ACKNOWLEDGE + response.getBody());
                } else {
                    log.debug(ACKNOWLEDGE);
                    oqmAckReplier.ack();
                }
                ackedNacked = true;

            } catch (Exception e) {
                log.debug(NOT_ACKNOWLEDGE, e);
            }

            if (!ackedNacked) oqmAckReplier.nack();
        };

        OqmSubscriber subscriber = OqmSubscriber.builder().subscription(subscription).messageReceiver(receiver).build();
        driver.subscribe(subscriber, destination);
        log.info("Just subscribed at topic {} subscription {} for tenant {}",
                subscription.getTopics().get(0), subscription.getName(), tenantInfo.getDataPartitionId());
    }

    private void registerControlTopicSubscriber(TenantInfo tenantInfo, OqmSubscription controlTopicSubscriber) {

        OqmMessageReceiver receiver = (oqmMessage, oqmAckReplier) -> {

            String pubsubMessage = oqmMessage.getData();
            Map<String, String> headerAttributes = oqmMessage.getAttributes();
            String subscriptionId = headerAttributes.get("subscription-id");
            String dataPartitionId = headerAttributes.get("data-partition-id");
            String topic = headerAttributes.get("topic");

            OqmSubscription subscription = OqmSubscription.builder()
                    .topic(OqmTopic.builder().name(topic).build())
                    .name(subscriptionId)
                    .build();

            registerSubscriber(tenantInfoFactory.getTenantInfo(dataPartitionId), subscription);

            log.debug(ACKNOWLEDGE);
            oqmAckReplier.ack();
        };

        OqmSubscriber subscriber = OqmSubscriber.builder().subscription(controlTopicSubscriber).messageReceiver(receiver).build();
        OqmDestination destination = getDestination(tenantInfo);
        driver.subscribe(subscriber, destination);

        log.info("Just subscribed at topic {} subscription {} for tenant {}",
                controlTopicSubscriber.getTopics().get(0), controlTopicSubscriber.getName(), tenantInfo.getDataPartitionId());
    }

    public List<OqmSubscription> getSubscriberableSubscriptions(TenantInfo tenantInfo) {
        return driver.listSubscriptions(null, INTERESTED_SUBSCRIPTIONS_QUERY, getDestination(tenantInfo));
    }

    private OqmDestination getDestination(TenantInfo tenantInfo) {
        return OqmDestination.builder().partitionId(tenantInfo.getDataPartitionId()).build();
    }

}
