/*
 *   Copyright 2017-2020, Schlumberger
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

package org.opengroup.osdu.notification.provider.gcp.pubsub.di;

import org.opengroup.osdu.core.common.cache.ICache;
import org.opengroup.osdu.core.common.cache.MultiTenantCache;
import org.opengroup.osdu.core.common.cache.VmCache;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.core.common.provider.interfaces.ITenantFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.stereotype.Component;

import java.util.Map;

@Component
@ConditionalOnProperty(name = "oqmDriver")
public class OqmSubscriptionCacheFactory {
    @Autowired
    private ITenantFactory tenantFactory;

    private MultiTenantCache<String> caches;

    public OqmSubscriptionCacheFactory(@Value("${app.expireTime}") int expireTime, @Value("${app.maxCacheSize}") int maxCacheSize) {
        this.caches = new MultiTenantCache<>(new VmCache<>(expireTime, maxCacheSize));
    }

    public void put(String key, String val, Map<String, String> headerAttributes) {
        this.partitionCache(headerAttributes).put(key, val);
    }

    public String get(String key, Map<String, String> headerAttributes) {
        return this.partitionCache(headerAttributes).get(key);
    }

    public void delete(String key, Map<String, String> headerAttributes) {
        this.partitionCache(headerAttributes).delete(key);
    }

    public void clearAll(Map<String, String> headerAttributes) {
        this.partitionCache(headerAttributes).clearAll();
    }

    private ICache<String, String> partitionCache(Map<String, String> headerAttributes) {
        String tenantId = headerAttributes.get(DpsHeaders.DATA_PARTITION_ID);
        TenantInfo tenantInfo = this.tenantFactory.getTenantInfo(tenantId);
        if (tenantInfo == null) {
            throw AppException.createUnauthorized(String.format("could not retrieve tenant info for data partition id: %s", tenantId));
        }
        return this.caches.get(String.format("%s:subscription", tenantInfo.getDataPartitionId()));
    }

}
